#include "hippo.h"
#include "debug.h"
#include <sys/time.h>
#include <string.h>
#include <assert.h>
#include <sys/mman.h>

#if !defined(_LTTNG_)
    #define tracepoint(...)
#else
    #include "hippo-tp.h"
/*    #define _COUNT_FIRE_TRANSITION_LOOP_ */
#endif

/*******************************************************************************
 * constants
 ******************************************************************************/
#ifndef _TICK_UNIT_
    #define _TICK_UNIT_ 1000000000
#endif
#define NOT_FIREABLE -1
#define FIREABLE 1
#define EVENT_PRIORITY 95
#define ENGINE_PRIORITY 99
#define PROCESS_PRIORITY 30
#define ALARM_IS_ACTIVE 1
#define ALARM_IS_NOT_ACTIVE 0
#define NO_PREVIOUS -1
#define NONE -1
#define ENABLE 1
#define NOT_ENABLE -1

/*******************************************************************************
 * macros
 ******************************************************************************/
#define errExit(msg)    do { perror(msg); exit(EXIT_FAILURE); \
                            } while (0)
#define REPORT_NULL_SYSCALL(call, string)	\
  do {int res_ = call; 				\
    if (res_ != 0) {				\
      fprintf(stderr,"%s: %s\n",string,strerror(res_));	\
      assert(res_ == 0);			\
    } 						\
  } while (0)

#define REPORT_MINUS_ONE_SYSCALL(call, string)	\
  do {int res_ = call; 				\
    if (res_ == -1) {				\
      fprintf(stderr,"%s: %s\n",string,strerror(res_));	\
      assert(res_ == -1);			\
    } 						\
  } while (0)
#define IS_TRANSITION_DELAYED(transition) (!(((transition)->minTime == 0)&&((transition)->maxTime ==0 || (transition)->maxTime ==-1)))
#define IS_TRANSITION_IMMEDIATE(transition) (((transition)->minTime == 0)&&((transition)->maxTime ==0))
#define IS_BOUND_TO_TASK(transition) ((transition)->call != NULL)
#define IS_BOUND_TO_EVENT(transition) ((transition)->event != NULL)
#define IS_TRANSITION_FIREABLE(transition) (setOfFireableTransitions[(transition)->idx].state == FIREABLE)
#define IS_TRANSITION_ENABLE(transition) (! (transition)->enabled_state)

#if defined XENOMAI
    #define UNBLOCK_ENGINE rt_sem_v(&semaphoreToCallTheEngine)
#elif defined POSIX
    #define UNBLOCK_ENGINE REPORT_MINUS_ONE_SYSCALL(sem_post(&semaphoreToCallTheEngine),"UNBLOCK_ENGINE sem_post")
#endif

#if defined XENOMAI
    #define BLOCK_ENGINE rt_sem_p(&semaphoreToCallTheEngine, 0)
#elif defined POSIX
  #define BLOCK_ENGINE  while (sem_wait(&semaphoreToCallTheEngine) == -1){ if (errno == EINTR) errno = 0; else {fprintf(stderr,"BLOCK_ENGINE: %s\n",strerror(errno)); assert(errno == EINTR); break;}}
    // TODO : find a more elegant solution with posix and avoid the test on EINTR
#endif

#if defined XENOMAI
    #define ACQUIRE_MUTEX_ENGINE rt_mutex_acquire_timed(&mutexDataEngine, NULL)
#elif defined POSIX
    #define ACQUIRE_MUTEX_ENGINE REPORT_NULL_SYSCALL(pthread_mutex_lock(&mutexDataEngine),"ACQUIRE_MUTEX_ENGINE")
#endif

#if defined XENOMAI
    #define RELEASE_MUTEX_ENGINE rt_mutex_release(&mutexDataEngine)
#elif defined POSIX
    #define RELEASE_MUTEX_ENGINE  REPORT_NULL_SYSCALL(pthread_mutex_unlock(&mutexDataEngine),"RELEASE_MUTEX_ENGINE")
#endif

#if defined XENOMAI
    #define SUSPEND_EXTERN_TASK(extern) rt_task_suspend(NULL)
#elif defined POSIX
    #define SUSPEND_EXTERN_TASK(extern) while (sem_wait((*((extern)->startTransition))->sem_suspend) == -1){ if (errno == EINTR) errno = 0; else {fprintf(stderr,"SUSPEND_EXTERN_TASK: %s\n",strerror(errno)); assert(errno == EINTR); break;}}
#endif

#if defined XENOMAI
    #define RESUME_EXTERN_TASK(transition) rt_task_resume(transition->task)
#elif defined POSIX
    #define RESUME_EXTERN_TASK(transition) REPORT_MINUS_ONE_SYSCALL(sem_post((transition)->sem_suspend),"RESUME_EXTERN_TASK")
#endif

/*******************************************************************************
 * private functions
 ******************************************************************************/
#if defined (XENOMAI)
    void * handle_global_clock(void *arg);
#elif defined (POSIX)
    static void  handle_global_clock(int sig, siginfo_t *si, void *uc);
#endif
void *handle_engine_execution(void *);
void *handle_execution_of_an_extern_function(void *arg);
void *handle_execution_of_an_extern_event(void *arg);
void start_engine();
void set_time_timer();
void compute_priority();
void initialize_structures_for_the_engine();
void spawn_task_engine();
void initialize_structures_for_global_clock();
void initialize_set_of_active_alarm();
void initialize_structures_for_extern_events_and_tasks();
void initialize_values_of_the_model();
void initialize_transitionWithPreSet();
void spawn_task_for_extern(Extern *ext);
void update_transition_initial_states();
void declare_extern_as_active(Extern *ext);
void activate_alarm_of_a_transition(Transition *tr);
void fire_transition(Transition *transition);
void add_token_to_place(Place *place);
void remove_transition_from_the_fireable_set(Transition *transition);
void add_transition_to_the_fireable_set(Transition *transition);
void remove_transition_from_the_enable_set(Transition *transition);
void add_transition_to_the_enable_set(Transition *transition);
void update_state_transition_with_pre_condition_isTrWithPre_isTrEnabled(Transition *tr);
void update_state_transition_with_pre_condition(Transition *transition);
void update_state_of_all_transitions_with_pre_conditions(void);
void update_state_of_transition_after_empty_a_previous_place(Transition *transition, int idx);
void update_state_of_transition_after_add_a_token_in_a_previous_place(Transition *transition, int idx);
void print_state(char * comment_char, bool is_text_color);
void print_marked_places();
void print_enabled_transitions();
void print_enabled_transitions_2();
void print_fireable_transitions();
void print_state_of_alarms();

/*******************************************************************************
 * extern variables
 ******************************************************************************/
extern Transition *listStartTransitions[NUMBER_OF_START_TRANSITIONS];
extern Transition *listEndTransitions[NUMBER_OF_END_TRANSITIONS];
extern Extern externset[NUMBER_OF_EXTERNS];
extern Place placeset[NUMBER_OF_PLACES];
extern Transition transitionset[NUMBER_OF_TRANSITIONS];
extern Edge edgeset[NUMBER_OF_EDGES];
extern fromPlaceToTransition preset[NUMBER_OF_PRE];
extern fromTransitionToPlace postset[NUMBER_OF_POST];
extern PriorityRelation setOfPriorityRelations[NUMBER_OF_PRIORITY_RELATIONS];

/*******************************************************************************
 * local variables
 ******************************************************************************/
#if defined (POSIX)    
    #define signum SIGRTMIN
#endif
unsigned long long int global_logical_time = 0;
int idxOfTheCurrentFireableTransition = NONE;
int idxOfTheLastFireableTransition = NONE;
int idxOfTheEarliestAlarm;
ActiveAlarm setOfActiveAlarms[NUMBER_OF_TRANSITIONS];
Extern *setOfFinishedExternEventsAndTasks[NUMBER_OF_EXTERNS];
int nbOfFinishedExterns = -1;
Fireable setOfFireableTransitions[NUMBER_OF_TRANSITIONS];
int idxOfTheCurrentEnableTransition = NONE;
Enable setOfEnableTransitions[NUMBER_OF_TRANSITIONS];
int priorityRelations[NUMBER_OF_TRANSITIONS][NUMBER_OF_TRANSITIONS];
#if !defined EXTERNAL_HIPPO_DEBUG
#if defined(_DEBUG_ON_)
    int hippo_debug = 1;
#else
    int hippo_debug = 0;
#endif
#endif
int hippo_shutdown = 0;
#if defined XENOMAI
    RT_ALARM clock_alarm;
    RT_TASK taskset[NUMBER_OF_TASKS];
    RT_SEM semaphoreToCallTheEngine;
    RT_MUTEX mutexDataEngine;
    RT_TASK taskEngine;
#elif defined POSIX
    timer_t clock_alarm;
    sigset_t mask;
    struct sigaction sa;
    struct sigevent sev;
    struct itimerspec its;
    pthread_t taskset[NUMBER_OF_TASKS];
    sem_t semaphoreToCallTheEngine;
    static pthread_mutex_t mutexDataEngine;
    pthread_t taskEngine;
#endif


/*******************************************************************************
 * function implementations
 ******************************************************************************/
void (*thread_starting_hook) () = NULL;

void call_thread_starting_hook(char *thread_name) //!OCLINT Ignore the parameter
{
  if (thread_starting_hook) {
    thread_starting_hook(NULL);	/* let portlib pick up a unique name */
  }
}

void print_scheduler(void)
{
   int schedType = sched_getscheduler(getpid());

   switch(schedType)
   {
     case SCHED_FIFO:
	   printf("Pthread Policy is SCHED_FIFO\n");
	   break;
     case SCHED_OTHER:
	   printf("Pthread Policy is SCHED_OTHER\n");
       break;
     case SCHED_RR:
	   printf("Pthread Policy is SCHED_OTHER\n");
	   break;
     default:
       printf("Pthread Policy is UNKNOWN\n");
   }
}

void initialize_lists(){
  for(int i = 0 ; i < NUMBER_OF_TRANSITIONS; i++){
    setOfFireableTransitions[i].state = NOT_FIREABLE;
    setOfFireableTransitions[i].prev  = NONE;
    setOfFireableTransitions[i].next  = NONE;

    setOfEnableTransitions[i].state = NOT_ENABLE;
    setOfEnableTransitions[i].prev  = NONE;
    setOfEnableTransitions[i].next  = NONE;

    setOfActiveAlarms[i].date = -1;
    setOfActiveAlarms[i].active = ALARM_IS_NOT_ACTIVE;
    setOfActiveAlarms[i].prev = NONE;
    setOfActiveAlarms[i].next = NONE;

    assert(transitionset[i].sizePreset < sizeof(ENABLE_STATE_TYPE) * 8
           || transitionset[i].sizePreset == sizeof(ENABLE_STATE_TYPE) * 8);
  }
}

#if defined (POSIX)
void block_sigalarm ()
{
  sigset_t   sigset;

  sigemptyset(&sigset);
  sigaddset(&sigset, signum + 1);
  REPORT_NULL_SYSCALL(pthread_sigmask(SIG_BLOCK, &sigset, NULL),"block_sigalarm");
}

void unblock_sigalarm ()
{
  sigset_t   sigset;

  sigemptyset(&sigset);
  sigaddset(&sigset, signum +1);
  REPORT_NULL_SYSCALL(pthread_sigmask(SIG_UNBLOCK, &sigset, NULL),"unblock_sigalarm");
}

void create_timer ()
{
    struct sigevent ReleaseEvent;
    ReleaseEvent.sigev_notify = 0; //SIGEV_SIGNAL;
    ReleaseEvent.sigev_signo = SIGRTMIN + 1;
    ReleaseEvent.sigev_value.sival_ptr = &clock_alarm;
    timer_create(CLOCK_REALTIME, &ReleaseEvent, &clock_alarm);

}
#endif

void engine_run() {
    mlockall(MCL_CURRENT | MCL_FUTURE);
    #if defined ACTIVATE_SCHED_FIFO
        struct sched_param rt_param;
        int rc;
        rt_param.sched_priority = sched_get_priority_max(SCHED_FIFO);
        rc=sched_setscheduler(getpid(), SCHED_FIFO, &rt_param);
        printf("test\n");
        if (rc){
            if (rc == EPERM){printf("No permission to set the scheduling policy\n"); perror(NULL); exit(-1);}
            else {printf("ERROR; pthread_create() rc is %d\n", rc); perror(NULL); exit(-1);}
        }
    #endif
    print_scheduler();

    printf("#             .''\n#   ._.-.___.' (`\\  >> Initialize engine structures <<\n"); // Art by Veronica Karlsson
    initialize_lists();
#if defined POSIX
    block_sigalarm();		/* this will block for everybody... */
#endif
    initialize_structures_for_the_engine();
    spawn_task_engine();
    initialize_structures_for_global_clock();
    printf("#  //(        ( `'  >> Engine is initialized and ready to start <<\n");
    printf("# '/ )\\ ).__. )     >> Compute initial state << \n");
    compute_priority();
    initialize_structures_for_extern_events_and_tasks();
    initialize_values_of_the_model();
    update_transition_initial_states();
    printf("# ' <' `\\ ._/'\\     >> GO hippo GO << \n#    `   \\     \\  vk\n");
    printf("# _TICK_UNIT_ = %d ns\n", _TICK_UNIT_);
#if defined NANOSLEEP
#elif defined POSIX
    create_timer();		/* no need to create it before */
#endif
    start_engine();
#if defined POSIX
    unblock_sigalarm();		/* this will unblock for main only... */
#endif
    while (pause() == -1 && errno == EINTR) {
      if (hippo_shutdown)
	     break;
      else
	     continue;
    }
    printf("# >> This is the end <<\n");
}

void initialize_structures_for_the_engine(){
    #if defined XENOMAI
        rt_sem_create(&semaphoreToCallTheEngine, "sync", 0, S_FIFO);
    #elif defined POSIX
        sem_init(&semaphoreToCallTheEngine, 0, 0);
    #endif
    #if defined XENOMAI
        rt_mutex_create(&mutexDataEngine, "mutex_engine");
    #elif defined POSIX
        pthread_mutexattr_t mattr;
        pthread_mutexattr_init(&mattr);
        pthread_mutexattr_setprotocol(&mattr, PTHREAD_PRIO_INHERIT);
        pthread_mutexattr_setpshared(&mattr, PTHREAD_PROCESS_SHARED);
        pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
        pthread_mutexattr_setrobust(&mattr, PTHREAD_MUTEX_ROBUST);
        pthread_mutex_init(&mutexDataEngine, &mattr);
    #endif
}

void initialize_values_of_the_model(){
    initial();
}

void spawn_task_engine(){
    #if defined XENOMAI
        rt_task_spawn(&taskEngine, "Engine", 0, ENGINE_PRIORITY, 0, handle_engine_execution, NULL);
    #elif defined POSIX
        #if !defined ACTIVATE_SCHED_FIFO
            pthread_attr_t attributs;
            pthread_attr_init(& attributs);
            pthread_create (&taskEngine, & attributs, handle_engine_execution, NULL);
        #else
            struct sched_param rt_param;
            pthread_attr_t rt_sched_attr;
            int rc;

            pthread_attr_init(& rt_sched_attr);
            pthread_attr_setinheritsched(&rt_sched_attr, PTHREAD_EXPLICIT_SCHED);
            pthread_attr_setschedpolicy(&rt_sched_attr, SCHED_FIFO);

            rt_param.sched_priority = sched_get_priority_max(SCHED_FIFO)-1; // ENGINE_PRIORITY
            pthread_attr_setschedparam(&rt_sched_attr, &rt_param);
            
            rc = pthread_create(&taskEngine, &rt_sched_attr, handle_engine_execution, NULL);
            if (rc){
                if (rc == EPERM){printf("No permission to set the scheduling policy\n"); perror(NULL); exit(-1);}
                else {printf("ERROR; pthread_create() rc is %d\n", rc); perror(NULL); exit(-1);}
            }
        #endif
    #endif
    #if defined AFFINITY
        cpu_set_t cpuset;
        CPU_ZERO(&cpuset);
        CPU_SET(0, &cpuset);
        #if defined POSIX
            pthread_setaffinity_np(taskEngine, sizeof(cpu_set_t), &cpuset);
        #elif defined XENOMAI
            rt_task_set_affinity(&taskEngine, &cpuset);
        #endif
    #endif
}

void register_alarm_handler ()
{
    struct sigaction ReleaseAction;
    ReleaseAction.sa_sigaction = handle_global_clock;
    sigemptyset(&ReleaseAction.sa_mask);
    ReleaseAction.sa_flags = SA_SIGINFO;
    sigaction(SIGRTMIN+1, &ReleaseAction, NULL); 
}

void initialize_structures_for_global_clock(){
    #if defined XENOMAI
      //rt_alarm_create(&clock_alarm,"alarm_global_clock", handle_global_clock,NULL);
    #elif defined POSIX
      register_alarm_handler();
    #endif
    initialize_set_of_active_alarm();
}

void initialize_set_of_active_alarm(){
    idxOfTheEarliestAlarm = NONE;
}

void update_transition_initial_states(){
    idxOfTheCurrentFireableTransition = NONE;
    for (Place *pl = &placeset[0]; pl < &placeset[NUMBER_OF_PLACES]; pl++) {
        if (pl->mark > 0) {
            for (Edge *e = pl->edge; e < pl->edge + pl->nbEdges; e++) {
                update_state_of_transition_after_add_a_token_in_a_previous_place(e->toTransition, e->posEnableState);
                log_printf("\x1B[31m%s %s %llu\x1B[0m\n", pl->name, e->toTransition->name, e->toTransition->enabled_state);
            }
        }
    }
}

void start_engine() {
    UNBLOCK_ENGINE;
#if defined NANOSLEEP
#else
    set_time_timer();
#endif
}

void set_time_timer(){
    #if defined (XENOMAI)
        //rt_alarm_start(&clock_alarm, 1, _TICK_UNIT_);
    #elif defined (POSIX)
        its.it_value.tv_sec = _TICK_UNIT_ / 1000000000;
        its.it_value.tv_nsec = _TICK_UNIT_ % 1000000000;
        its.it_interval.tv_sec = its.it_value.tv_sec;
        its.it_interval.tv_nsec = its.it_value.tv_nsec;
        if (timer_settime(clock_alarm, 0, &its, NULL) == -1)
            errExit("timer_settime");
    #endif
}

void compute_priority(){
    for (int i=0; i < NUMBER_OF_TRANSITIONS; i++){
        transitionset[i].priority = 0;
        transitionset[i].isHigherPriority = 0;
        for (int j=0; j < NUMBER_OF_TRANSITIONS; j++){
            priorityRelations[i][j] = 0;
        }
    }

    for (int i = 0; i < NUMBER_OF_PRIORITY_RELATIONS; i ++){
        int l = (setOfPriorityRelations[i].lower)->idx;
        int h = (setOfPriorityRelations[i].higher)->idx;
        priorityRelations[l][h] = 1;
        transitionset[l].priority = 1;
    }
    int test = 1;

    while(test){
        test = 0;
        for (int i=0; i < NUMBER_OF_TRANSITIONS; i++){
            for (int j=0; j < NUMBER_OF_TRANSITIONS; j++){
                if (priorityRelations[i][j] == 1){
                    for (int k = 0; k < NUMBER_OF_TRANSITIONS; k++){
                        if ((priorityRelations[j][k] == 1)&&(priorityRelations[i][k] != 1)){
                            priorityRelations[i][k]= 1;
                            test = 1;
                        }
                    }
                }
            }
        }
    }

    for (int i=0; i < NUMBER_OF_TRANSITIONS; i++){
        for (int j=0; j < NUMBER_OF_TRANSITIONS; j++){
            if (priorityRelations[i][j] == 1){
                transitionset[j].isHigherPriority = 1;
            }
        }
    }
}

void initialize_structures_for_extern_events_and_tasks() {
    for (Transition *tr = &transitionset[0]; tr < &transitionset[NUMBER_OF_TRANSITIONS]; tr++) {
        if (tr->minTime != 0) {
            tr->alarm = 1;
            log_printf("\x1B[31mAlarm handler %s is created\x1B[0m\n", tr->name);
        }
        for (int i = 0; i < tr->sizePreset; i++) {
            tr->enabled_state |= (1 << i);
        }
        if IS_TRANSITION_ENABLE(tr) {
          add_transition_to_the_enable_set(tr);
        }
    }
    for (Extern *ext = &externset[0]; ext < &externset[NUMBER_OF_EXTERNS]; ext++) {
        spawn_task_for_extern(ext);
    }
}

void spawn_task_for_extern(Extern *ext){
        if (ext->isEvent) {
            #if defined XENOMAI
                rt_task_spawn(&taskset[ext->idx], ext->name, 0, EVENT_PRIORITY, 0, handle_execution_of_an_extern_event, ext);
            #elif defined POSIX                
                sem_t * sem_id= malloc(sizeof(sem_t));
                sem_init(sem_id, 0, 0);
                for (int i =0; i < ext->nbStartTransition; i++){
                    Transition *tr = *(ext->startTransition + i);
                    tr->sem_suspend = sem_id;
                }
                #if !defined ACTIVATE_SCHED_FIFO                    
                    pthread_attr_t attributs;
                    pthread_attr_init(& attributs);
                    pthread_create (&taskset[ext->idx], & attributs, handle_execution_of_an_extern_event, ext);
                #else
                    struct sched_param rt_param;
                    pthread_attr_t rt_sched_attr;
                    int rc;
                    pthread_attr_init(& rt_sched_attr);
                    pthread_attr_setinheritsched(&rt_sched_attr, PTHREAD_EXPLICIT_SCHED);
                    pthread_attr_setschedpolicy(&rt_sched_attr, SCHED_FIFO);
                    rt_param.sched_priority = sched_get_priority_max(SCHED_FIFO)-10;
                    pthread_attr_setschedparam(&rt_sched_attr, &rt_param);
                    rc = pthread_create(&taskset[ext->idx], &rt_sched_attr, handle_execution_of_an_extern_event, ext);
                    if (rc){printf("ERROR; pthread_create() rc is %d\n", rc); perror(NULL); exit(-1);}
                #endif
            #endif
            log_printf("\x1B[Task for extern event [%s] is spawned\x1B[0m\n", ext->name);
        } else {
            #if defined XENOMAI
                rt_task_spawn(&taskset[ext->idx], ext->name, 0, PROCESS_PRIORITY, 0, handle_execution_of_an_extern_function, ext);
            #elif defined POSIX
                sem_t * sem_id= malloc(sizeof(sem_t));
                sem_init(sem_id, 0, 0);
                for (int i =0; i < ext->nbStartTransition; i++){
                    Transition *tr = *(ext->startTransition + i);
                    tr->sem_suspend = sem_id;
                }
                #if !defined ACTIVATE_SCHED_FIFO                    
                    pthread_attr_t attributs;
                    pthread_attr_init(& attributs);
                    pthread_create (&taskset[ext->idx], & attributs, handle_execution_of_an_extern_function, ext);
                #else
                    struct sched_param rt_param;
                    pthread_attr_t rt_sched_attr;
                    int rc;
                    pthread_attr_init(& rt_sched_attr);
                    pthread_attr_setinheritsched(&rt_sched_attr, PTHREAD_EXPLICIT_SCHED);
                    pthread_attr_setschedpolicy(&rt_sched_attr, SCHED_FIFO);
                    rt_param.sched_priority = sched_get_priority_max(SCHED_FIFO)-9;
                    pthread_attr_setschedparam(&rt_sched_attr, &rt_param);
                    rc = pthread_create(&taskset[ext->idx], &rt_sched_attr, handle_execution_of_an_extern_function, ext);
                    if (rc){printf("ERROR; pthread_create() rc is %d\n", rc); perror(NULL); exit(-1);}
                #endif
            #endif
            (*(ext->startTransition))->task = &taskset[ext->idx];
            log_printf("\x1B[31mTask for externe function [%s] is spawned\x1B[0m\n", ext->name);
        }

#if defined AFFINITY
        cpu_set_t cpuset;
        long number_of_processors = sysconf(_SC_NPROCESSORS_ONLN);
        CPU_ZERO(&cpuset);
        for (int cpu = 1; cpu < number_of_processors; cpu++) 
            CPU_SET(cpu, &cpuset);
         #if defined POSIX
            pthread_setaffinity_np(taskset[ext->idx], sizeof(cpu_set_t), &cpuset);
        #elif defined XENOMAI
            rt_task_set_affinity(&taskset[ext->idx], &cpuset);
        #endif
#endif
}

#if defined (XENOMAI)
void * handle_global_clock(void *arg)
#elif defined (POSIX)
static void  handle_global_clock(int sig, siginfo_t *si, void *uc) //!OCLINT parameters is unsued here
#endif
{
    // TODO : vérifier que la priorité du handler et de la tâche
    // de l'engine est plus élévée que les
    // tâches de externs (event et task) => risque d'incohérence entre appel
    // engine et execution de l'engine

    UNBLOCK_ENGINE; // Async-Signal Safety
}

int engine_one_turn(int idxOfExternsToDealWith, Extern * (*t)[]){
  tracepoint(hippo, start_one_turn_engine, global_logical_time);
  int idx = idxOfTheEarliestAlarm;
#if defined(_COUNT_FIRE_TRANSITION_LOOP_)
  int loop_cnt = 0;
#endif
  while ((idx != NONE) && (setOfActiveAlarms[idx].date == global_logical_time)){
      Transition *tr = &transitionset[idx];
      if IS_TRANSITION_ENABLE(tr) {
          add_transition_to_the_fireable_set(tr);
      }
      idx = setOfActiveAlarms[idx].next;
  }
  while (idxOfExternsToDealWith >= 0){
      Extern *ext = (*t)[idxOfExternsToDealWith];
      if (ext->endTransition != 0) {
          *(ext->terminate) = 1;
          for (Transition **tr_idx = ext->endTransition ; tr_idx < (ext->endTransition + ext->nbEndTransition) ; tr_idx++) {
              update_state_transition_with_pre_condition((*(tr_idx)));
          }
      }
      *(ext->ready) = 1;
      for (Transition **tr_idx = ext->startTransition ; tr_idx < (ext->startTransition + ext->nbStartTransition) ; tr_idx++) {
          update_state_transition_with_pre_condition((*(tr_idx)));
      }
      idxOfExternsToDealWith --;
  }
  while (idxOfTheCurrentFireableTransition != NONE) {
      log_state;
      fire_transition(&transitionset[idxOfTheCurrentFireableTransition]);
#if defined(_COUNT_FIRE_TRANSITION_LOOP_)
      loop_cnt++;
#endif
  }
  tracepoint(hippo, end_one_turn_engine_cnt, global_logical_time, loop_cnt);
  return idxOfExternsToDealWith;
}

void *handle_engine_execution(void *arg) { //!OCLINT parameter is unused 
    call_thread_starting_hook("task_engine");
    Extern *localCpyOfActiveExterns[NUMBER_OF_EXTERNS] = { NULL };
    int idxOfExternsToDealWith = -1;
    BLOCK_ENGINE;
#if defined (XENOMAI)
    rt_task_set_periodic(NULL, TM_NOW, _TICK_UNIT_);
#endif
    ACQUIRE_MUTEX_ENGINE;
    idxOfExternsToDealWith = engine_one_turn(idxOfExternsToDealWith, &setOfFinishedExternEventsAndTasks);
    RELEASE_MUTEX_ENGINE;
#if defined NANOSLEEP
    struct timespec next_time;
    clock_gettime(CLOCK_REALTIME, &next_time); 
#endif
    while (1) {
#if defined (XENOMAI)
        rt_task_wait_period(NULL);
#elif defined NANOSLEEP
        next_time.tv_nsec = next_time.tv_nsec + _TICK_UNIT_ % 1000000000;
        if(next_time.tv_nsec >= 1000000000) {
            next_time.tv_nsec -= 1000000000;
            next_time.tv_sec += 1;
        }
        next_time.tv_sec += _TICK_UNIT_ / 1000000000;
        clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &next_time, NULL);
#else
        BLOCK_ENGINE;
#endif
        bool is_an_alarm_ready = 0;
        bool is_an_extern_ready = 0;
        global_logical_time ++;
        is_an_alarm_ready = (setOfActiveAlarms[idxOfTheEarliestAlarm].date == global_logical_time);
        ACQUIRE_MUTEX_ENGINE;
        if (nbOfFinishedExterns >= 0){
            is_an_extern_ready = 1;
            memcpy(&localCpyOfActiveExterns, &setOfFinishedExternEventsAndTasks, (nbOfFinishedExterns+1)*sizeof(Extern *));
            idxOfExternsToDealWith = nbOfFinishedExterns;
            nbOfFinishedExterns = -1;
        }
        RELEASE_MUTEX_ENGINE;

        tracepoint(hippo, clock, global_logical_time, is_an_alarm_ready, is_an_extern_ready );
        log_time;

        if (is_an_alarm_ready || is_an_extern_ready){
            idxOfExternsToDealWith = engine_one_turn(idxOfExternsToDealWith, &localCpyOfActiveExterns);
        }
    }
}

void *handle_execution_of_an_extern_function(void *arg) {
    Extern *ext = (Extern *) arg;
    call_thread_starting_hook("call_handler");
    while (1) {
        SUSPEND_EXTERN_TASK(ext);
        log_printf("\x1B[34m@%llu: The external function [%s] is called\x1B[0m\n", global_logical_time, ext->name);
        (ext->function)();
        log_printf("\x1B[34m@%llu: The external function [%s] is finished\x1B[0m\n", global_logical_time, ext->name);
        declare_extern_as_active(ext);
    }
}

void *handle_execution_of_an_extern_event(void *arg) {
    Extern *ext = (Extern *) arg;
    call_thread_starting_hook("event_handler");
    while (1) {
        log_printf("\x1B[34m@%llu: The external event [%s] listens\x1B[0m\n", global_logical_time, ext->name);
        (ext->function)();
        log_printf("\x1B[34m@%llu: The external event [%s] is occurred\x1B[0m\n", global_logical_time, ext->name);
        declare_extern_as_active(ext);
    }
}

void declare_extern_as_active(Extern *ext){
    ACQUIRE_MUTEX_ENGINE;
    nbOfFinishedExterns++;
    setOfFinishedExternEventsAndTasks[nbOfFinishedExterns] = ext;
    RELEASE_MUTEX_ENGINE;
}

void activate_alarm_of_a_transition(Transition *tr){
    setOfActiveAlarms[tr->idx].date = global_logical_time + tr->minTime;
    setOfActiveAlarms[tr->idx].active = ALARM_IS_ACTIVE;
    if (idxOfTheEarliestAlarm == NONE){
        idxOfTheEarliestAlarm = tr->idx;
        setOfActiveAlarms[tr->idx].next = NONE;
        setOfActiveAlarms[tr->idx].prev = NONE;
    } else {
        int next_idx = idxOfTheEarliestAlarm;
        int prev_idx = setOfActiveAlarms[next_idx].prev;
        while (setOfActiveAlarms[next_idx].date < setOfActiveAlarms[tr->idx].date){
            prev_idx = next_idx;
            next_idx = setOfActiveAlarms[next_idx].next;
            if (next_idx == NONE)
                break;
        }
        setOfActiveAlarms[tr->idx].next = next_idx;
        setOfActiveAlarms[tr->idx].prev = prev_idx;
        if (prev_idx == NONE){
            idxOfTheEarliestAlarm = tr->idx;
        } else{
            setOfActiveAlarms[prev_idx].next = tr->idx;
        }
        if (next_idx != NONE){
            setOfActiveAlarms[next_idx].prev = tr->idx;
        }
    }
}

void __attribute__((optimize("O0"))) remove_alarm(Transition *tr){
    int idx_next = setOfActiveAlarms[tr->idx].next;
    int idx_prev = setOfActiveAlarms[tr->idx].prev;
    setOfActiveAlarms[tr->idx].active = ALARM_IS_NOT_ACTIVE;
    setOfActiveAlarms[tr->idx].prev = NONE;
    setOfActiveAlarms[tr->idx].next = NONE;
    setOfActiveAlarms[tr->idx].date = -1;
    if (idxOfTheEarliestAlarm == tr->idx) /* May be we are not removing the earliest one... */
      idxOfTheEarliestAlarm = idx_next;	  /* Just set it if we removed the earliest alarm. */
    if (idx_prev == NONE){
        setOfActiveAlarms[idx_next].prev = NONE;
    } else {
        setOfActiveAlarms[idx_prev].next = idx_next;
    }
    if (idx_next == NONE){
        setOfActiveAlarms[idx_prev].next = NONE;
    } else {
      setOfActiveAlarms[idx_next].prev = idx_prev;
    }
}

void __attribute__((optimize("O0")))remove_transition_from_the_fireable_set(Transition *tr) {
    int idx = tr->idx;
    if (idxOfTheLastFireableTransition == idx)
      idxOfTheLastFireableTransition = setOfFireableTransitions[idx].prev;
    if (setOfFireableTransitions[idx].prev != NONE) {
        setOfFireableTransitions[setOfFireableTransitions[idx].prev].next = setOfFireableTransitions[idx].next;
    } else {
        idxOfTheCurrentFireableTransition = setOfFireableTransitions[idx].next;
    }
    if (setOfFireableTransitions[idx].next != NONE)
      setOfFireableTransitions[setOfFireableTransitions[idx].next].prev = setOfFireableTransitions[idx].prev;

    setOfFireableTransitions[idx].state = NOT_FIREABLE;
    setOfFireableTransitions[idx].prev = NONE;
    setOfFireableTransitions[idx].next = NONE;
}

void __attribute__((optimize("O0"))) add_transition_to_the_fireable_set(Transition *transition) {
    int transitionIdx = transition->idx;
    if (transition->priority){
        for (int i = 0; i < NUMBER_OF_TRANSITIONS ; i++){
            if ((priorityRelations[transition->idx][i]) && (setOfFireableTransitions[i].state == FIREABLE)){
                return;
            }
        }
    }
    setOfFireableTransitions[transitionIdx].state = FIREABLE;
    setOfFireableTransitions[transitionIdx].next = NONE;
    if (idxOfTheCurrentFireableTransition == NONE) { //empty set
        idxOfTheCurrentFireableTransition = transitionIdx;
        setOfFireableTransitions[transitionIdx].prev = NONE;
    }
    else {
        setOfFireableTransitions[transitionIdx].prev = idxOfTheLastFireableTransition;
        setOfFireableTransitions[idxOfTheLastFireableTransition].next = transitionIdx;
    }
    idxOfTheLastFireableTransition = transitionIdx;

    if (transition->isHigherPriority){
        for (int i = 0; i < NUMBER_OF_TRANSITIONS ; i++){
            if ((priorityRelations[i][transition->idx]) && (setOfFireableTransitions[i].state == FIREABLE)){
                remove_transition_from_the_fireable_set(&transitionset[i]);
            }
        }
    }
}

void __attribute__((optimize("O0"))) remove_transition_from_the_enable_set(Transition *tr){
  int idx = tr->idx;
  if(setOfEnableTransitions[idx].state == NOT_ENABLE)
    return;
  if (setOfEnableTransitions[idx].prev == NONE){//list head
    idxOfTheCurrentEnableTransition = setOfEnableTransitions[idx].next;
  }
  else{
    setOfEnableTransitions[setOfEnableTransitions[idx].prev].next = setOfEnableTransitions[idx].next;
  }
  if (setOfEnableTransitions[idx].next != NONE)
    setOfEnableTransitions[setOfEnableTransitions[idx].next].prev = setOfEnableTransitions[idx].prev;

  setOfEnableTransitions[idx].state = NOT_ENABLE;
  setOfEnableTransitions[idx].prev = NONE;
  setOfEnableTransitions[idx].next = NONE;
}

void __attribute__((optimize("O0"))) add_transition_to_the_enable_set(Transition *tr){
  int idx = tr->idx;
  if(setOfEnableTransitions[idx].state == ENABLE)
    return;
  setOfEnableTransitions[idx].state = ENABLE;
  setOfEnableTransitions[idx].prev = NONE;
  setOfEnableTransitions[idx].next = idxOfTheCurrentEnableTransition;
  if (idxOfTheCurrentEnableTransition != NONE)
    setOfEnableTransitions[idxOfTheCurrentEnableTransition].prev = idx;
  idxOfTheCurrentEnableTransition = idx;
}

void update_state_of_transition_after_remove_a_token(Transition *tr, int idxOfTheArcFromPreviousPlace) {
    if IS_TRANSITION_ENABLE(tr) {
        if IS_TRANSITION_DELAYED(tr) {
            log_printf("\x1B[34m@%llu: The alarm [%s] is stoped\x1B[0m\n", global_logical_time, tr->name);
            remove_alarm(tr);
        }
        if IS_TRANSITION_FIREABLE(tr) {
            remove_transition_from_the_fireable_set(tr);
        }
    }
    tr->enabled_state = tr->enabled_state | (1 << idxOfTheArcFromPreviousPlace);
    if (!IS_TRANSITION_ENABLE(tr)) {
      remove_transition_from_the_enable_set(tr);
    }
}

void update_state_transition_with_pre_condition_isTrWithPre_isTrEnabled(Transition *tr) {
  //reorder condition evaluation as bellow :
  // A : tr->alarm, F : IS_TRANSITION_FIREABLE, P : tr->pre()
  //                    -  F  --- not P
  //        -  not A  -
  //      -            -  not F - P
  // ---
  //      -            -  P
  //        -  A  ----
  //                   -  not P
    if (tr->alarm == 0) {//not A
      bool p = tr->pre();
      if IS_TRANSITION_FIREABLE(tr) {//F
        if(!p){//not P
          remove_transition_from_the_fireable_set(tr);
        }
      }
      else {//not F
        if(p) {//P
          add_transition_to_the_fireable_set(tr);
        }
      }
    }
    else {//A
      bool p = tr->pre();
      if(p) {//P
        if (setOfActiveAlarms[tr->idx].active == ALARM_IS_NOT_ACTIVE){
            activate_alarm_of_a_transition(tr);
            log_printf("\x1B[34m@%llu: The alarm [%s] is started (expired in %d)\x1B[0m\n", global_logical_time, tr->name, tr->minTime);
        }
      }
      else {//not P
        if (setOfActiveAlarms[tr->idx].active == ALARM_IS_ACTIVE){
            remove_alarm(tr);
            log_printf("\x1B[34m@%llu: The alarm [%s] is stopped\x1B[0m\n", global_logical_time, tr->name);
        }
      }
    }
}

void update_state_transition_with_pre_condition(Transition *tr) {
    if (IS_TRANSITION_ENABLE(tr) &&  tr->pre != NULL){
      update_state_transition_with_pre_condition_isTrWithPre_isTrEnabled(tr);
    }
}

void update_state_of_all_transitions_with_pre_conditions() {
    Transition *tr;
    int idx = idxOfTheCurrentEnableTransition;
    while (idx != NONE) {
        tr = &transitionset[idx];
        if(tr->pre != NULL)
          update_state_transition_with_pre_condition_isTrWithPre_isTrEnabled(tr);
        idx = setOfEnableTransitions[idx].next;
    }
}

void update_state_of_transition_after_add_a_token_in_a_previous_place(Transition *tr, int idxOfTheArcFromPreviousPlace) {
    int previously = IS_TRANSITION_ENABLE(tr); //!OCLINT useless parentheses du to macro
    tr->enabled_state = tr->enabled_state & ~(1 << idxOfTheArcFromPreviousPlace);
    if IS_TRANSITION_ENABLE(tr) {
      add_transition_to_the_enable_set(tr);
    }
    if ( IS_TRANSITION_ENABLE(tr) && !previously) {
        if IS_TRANSITION_IMMEDIATE(tr) {
            if ((tr->pre == 0) || (tr->pre())) {
                add_transition_to_the_fireable_set(tr);
            }
        } else {
            if (tr->alarm) {
                if (tr->pre != NULL) {
                    if (tr->pre() == 1) {
                        activate_alarm_of_a_transition(tr);
                        log_printf("\x1B[34m@%llu: The alarm [%s] is started (expired in %d)\x1B[0m\n", global_logical_time, tr->name, tr->minTime);
                    }
                } else {
                    activate_alarm_of_a_transition(tr);
                    log_printf("\x1B[34m@%llu: The alarm [%s] is started (expired in %d)\x1B[0m\n", global_logical_time, tr->name, tr->minTime);
                }
            }
        }
    }
}

void fire_transition(Transition *tr) {
    log_fire(global_logical_time, tr);
    remove_transition_from_the_fireable_set(tr);
    for (fromPlaceToTransition *arc = tr->preset; arc < tr->preset + tr->sizePreset; arc++) {
        if (!arc->read){ // The arc is not a read arc
            arc->fromPlace->mark--;
            if (arc->fromPlace->mark == 0){
                for (Edge *e = arc->fromPlace->edge ; e < (arc->fromPlace->edge + arc->fromPlace->nbEdges) ; e++) {
                    update_state_of_transition_after_remove_a_token(e->toTransition, e->posEnableState);
                }
            }
        }
    }
    /* Execute the action of the transition */
    if (tr->act) {
        tr->act();
        update_state_of_all_transitions_with_pre_conditions();
        //TODO : update only transitions with pre condition changed by the act
    }
    for (fromTransitionToPlace *post = tr->postset ; post < (tr->postset + tr->sizePostset) ; post++) {
        add_token_to_place(post->toPlace);
    }
    if IS_BOUND_TO_TASK(tr) {
        RESUME_EXTERN_TASK(tr);
    }
    log_state;
}

void add_token_to_place(Place * pl) {
    pl->mark++;
    if (pl->mark == 1) {
        for (Edge *e = pl->edge ; e < (pl->edge + pl->nbEdges) ; e++) {
          update_state_of_transition_after_add_a_token_in_a_previous_place(e->toTransition, e->posEnableState);
        }
    }
}

void print_marked_places(){
    char *separateur = ", ";
    printf("| Marked@%llu: [", global_logical_time);
    for (Place *pl = &placeset[0]; pl < &placeset[NUMBER_OF_PLACES]; pl++) {
        if (pl == &placeset[NUMBER_OF_PLACES - 1])
            separateur = "";
        if (pl->mark)
            printf("%s: %d%s", pl->name, pl->mark, separateur);
    }
    printf("]");
}

void print_enabled_transitions(){
    char *separateur = ", ";
    printf("| Enabled@%llu: [", global_logical_time);
    for (Transition *tr = &transitionset[0]; tr < &transitionset[NUMBER_OF_TRANSITIONS]; tr++) {
        if (tr == &transitionset[NUMBER_OF_TRANSITIONS - 1]){
            separateur = "";
        }
        if IS_TRANSITION_ENABLE(tr) {
          if ( ((tr->pre) && (tr->pre()))
               || (! tr->pre) )
          {
              printf("%s [%llu] %s ", tr->name, IS_TRANSITION_IMMEDIATE(tr) ? 0 : setOfActiveAlarms[tr->idx].date - global_logical_time, separateur);
          }
        }
    }
    printf("]");
}

void print_enabled_transitions_2(){
    Transition *tr;
    int idx = idxOfTheCurrentEnableTransition;
    char *separateur = ", ";
    printf("| Enabled@%llu: [", global_logical_time);
    while (idx != NONE) {
        tr = &transitionset[idx];
        if (setOfEnableTransitions[idx].next == NONE) {
            separateur = "";
        }
        if IS_TRANSITION_ENABLE(tr) {
            if ( ((tr->pre) && (tr->pre()))
                 || (! tr->pre) )
            {
                printf("%s [%llu] %s ", tr->name, IS_TRANSITION_IMMEDIATE(tr) ? 0 : setOfActiveAlarms[tr->idx].date - global_logical_time, separateur);
            }
        }
        idx = setOfEnableTransitions[idx].next;
    }
    printf("]");
}

void print_fireable_transitions(){
    char *separateur = ", ";
    printf("| Fireable@%llu: [", global_logical_time);
    int b = idxOfTheCurrentFireableTransition;
    while (b != NONE) {
        if (setOfFireableTransitions[b].next == NONE)
            separateur = "";
        printf("%s%s", transitionset[b].name, separateur);
        b = setOfFireableTransitions[b].next;
    }
    printf("]");
}

void print_state(char * comment_char, bool is_text_color) {
    char *color_begin = "";
    char *color_end = "";

    if (is_text_color){
        color_begin = "\x1B[35m";
        color_end = "\x1B[0m";
    }
    printf("%s%s\t", color_begin, comment_char);
    print_marked_places();
    printf("\n");
    printf("%s\t", comment_char);
    // print_enabled_transitions();
    print_enabled_transitions_2();
    printf("\n");
    printf("%s\t", comment_char);
    print_fireable_transitions();
    printf("%s\n", color_end);
}

void print_state_of_alarms(){
    for (int i = 0; i < NUMBER_OF_TRANSITIONS; i++)
        printf("{%d: nxt:%d, prv:%d, dt:%llu, ctv:%d}\n", i, setOfActiveAlarms[i].next, setOfActiveAlarms[i].prev, setOfActiveAlarms[i].date, setOfActiveAlarms[i].active);
    printf("\n");
}
